import math
from numbers import Number


def upRange(start, stop, step):
  while start <= stop:
    yield start
    start += abs(step)


def downRange(start, stop, step):
  while start >= stop:
    yield start
    start -= abs(step)


def math_isPrime(n):
    # https://en.wikipedia.org/wiki/Primality_test#Naive_methods
    # If n is not a number but a string, try parsing it.
    if not isinstance(n, Number):
      try:
        n = float(n)
      except:
        return False
    if n == 2 or n == 3:
      return True
    # False if n is negative, is 1, or not whole, or if n is divisible by 2 or 3.
    if n <= 1 or n % 1 != 0 or n % 2 == 0 or n % 3 == 0:
      return False
    # Check all the numbers of form 6k +/- 1, up to sqrt(n).
    for x in range(6, int(math.sqrt(n)) + 2, 6):
      if n % (x - 1) == 0 or n % (x + 1) == 0:
        return False
    return True


def PrimoSuccessivo(x):
    """
      Determina il numero primo successivo al numero x specificato.
      Funziona anche se viene passato un numero negativo
      o nullo (nel qual caso il primo successivo è due).
      Se il numero passato non è intero lo trasforma in intero arrotondando per difetto.
        Usa: ...
        Parametri:
          x, numero naturale
        Valore restituito:
          i numero primo successivo ad x
    """

    i = math.floor(x) + 1
    # Controlla via via se il numero successivo è un
    # numero primo, se è così il ciclo viene interrotto.
    while not math_isPrime(i):
      i = (i if isinstance(i, Number) else 0) + 1
    return i


def ScomporreInFattori(n):
    """
        Scompone in fattori primi un numero n specificato.
        Se n è nullo o non è intero ritorna una lista vuota,
        se n è negativo ritorna i fattori primi di -n
            Usa: ...
            Parametri:
                n, numero naturale specificato
            Valore restituito:
                una lista contenente i fattori di n, eventualmente ripetuti
    """

    # Se n è nullo o se n non è un numero intero restituisce una lista nulla.
    if n == 0 or not n % 1 == 0:
        return []
    # Se n è un intero negativo ne cambia il segno.
    if n < 0:
        n = -1 * n
    l = []
    p = 2
    # Ripete il ciclo controllando se il numero primo successivo
    # è un divisore di n, se è così aggiunge il numero
    # primo ai divisori e divide n per il numero trovato.
    # Così via fino a quando n non diventa uno (e quindi non ha più divisori).
    while not n == 1:
        # Se il resto della divisione è uguale a 0, p è un fattore di n
        if n % p == 0:
            l.append(p)
            n = n / p
        else:
            p = PrimoSuccessivo(p)
    return l


def Fattoriale(x):
  """
    Calcola il fattoriale del numero specificato (che deve
    essere un numero naturale) calcolando il prodotto
    dei numeri naturali da 2 al numero specificato.
    In caso il numero specificato sia 0 o 1 la funzione restituisce 1.
    Usa: ...
    Parametri:
      x un numero naturale
    Valore restituito:
      il fattoriale di x se x è naturale, infinito in caso contrario
  """

  # Caso particolare: 0! è uguale a uno per definizione.
  if x == 0:
    return 1
  # Caso particolare: 1! è uguale a uno.
  if x == 1:
    return 1
  # Se il numero specificato è negativo restituisce infinito
  if x < 0:
    return float('inf')
  # Se il numero specificato non è un intero restituisce infinito (N.B.:
  # a questo punto il numero specificato può essere solo positivo)
  if not x % 1 == 0:
    return float('inf')
  p = 1
  for i in (2 <= float(x)) and upRange(2, float(x), 1) or downRange(2, float(x), 1):
    p = p * i
  return p


def Newton_e(iterazioni):
    """
        Utilizza il metodo di Newton per approssimare e,
        tante volte quante il numero di iterazioni specificato.
        Approssima una soluzione dell'equazione ln(x)-1=0.
        Usa: ...
        Parametri:
            iterazioni
        Valore restituito:
            x_0, numero decimaleche approssima e
    """

    x_0 = 2
    # A ogni iterazione il valore di x_0 è l'ascissa del
    # punto di intersezione della tangente alla curva
    # (nel precedente x_0) con l'asse delle ascisse.
    for count in range(int(iterazioni)):
        x_0 = x_0 - (math.log(x_0) - 1) * x_0
    return x_0


def Newton_r2(iterazioni):
    """
        Utilizza il metodo di Newton per approssimare radice di 2.
        Determina una soluzione approssimata dell'equazione x^2-2=0.
        Usa: ...
        Parametri:
            iterazioni
        Valore restituito:
            x_0, numero che approssima radice di 2
    """

    x_0 = 1
    # A ogni iterazione il valore di x_0 è l'ascissa del
    # punto di intersezione della tangente alla curva
    # (nel precedente x_0) con l'asse delle ascisse.
    for count2 in range(int(iterazioni)):
        x_0 = x_0 - (x_0 ** 2 - 2) / (2 * x_0)
    return x_0


def Newton_phi(iterazioni):
    """
      Utilizza il metodo di Newton per approssimare radice di
      il valore della sezione aurea phi. Determina una soluzione
      approssimata dell'equazione x^2-5=0 e successivamente calcola
      il valore della sezione aurea utilizzando la soluzione trovata.
                  Usa: ...
                  Parametri:
                      iterazioni
                  Valore restituito:
                      (1+x0)/2, approssimazione della sezione aurea
    """

    x_0 = 1
    # A ogni iterazione il valore di x_0 è l'ascissa del
    # punto di intersezione della tangente alla curva
    # (nel precedente x_0) con l'asse delle ascisse.
    for count3 in range(int(iterazioni)):
      x_0 = x_0 - (x_0 ** 2 - 5) / (2 * x_0)
    return (1 + x_0) / 2


def MinimoScostamento(ripetizioni, trascendente):
    """
        Calcola l'approssimazione di un numero irrazionale o trascendente con il metodo del minimo scostamento.
        A ogni iterazione moltiplica l'approssimazione per il rapporto dei due primi successivi in modo
        da ridurre la differenza con il numero da approssimare. Le coppie di primi
        sono determinate dalla funzione MinimoScostamentoLista
        Se trascendente è nullo, ritorna infinito.
        Usa: MinimoScostamentoLista
        Parametri:
            ripetizioni, numero di iterazioni
            trascendente, numero irrazionale o trascendente da approssimare
        Valore restituito:
            un numero decimale dato dal prodotto di razionali che approssima il numero
    """

    x = MinimoScostamentoLista(ripetizioni, trascendente)
    if not len(x):
      return float('inf')
    # Il primo elemento della lista è il segno del numero da approssimare
    t = x.pop(0)
    while not not len(x):
      t = t * x.pop(0)
      t = t / x.pop(0)
    return t


def MinimoScostamentoLista(ripetizioni, trascendente):
    """
      Calcola l'approssimazione di un numero irrazionale o trascendente con il metodo del minimo scostamento.
      A ogni iterazione moltiplica l'approssimazione per il rapporto dei due primi successivi in modo
      da ridurre la differenza con il numero da approssimare. Lo sviluppo è memorizzato in una lista
      in cui il primo elemento è il segno, gli elementi successivi, a coppie, individuano
      un numero razionale; l'approssimazione si può ottenere moltiplicando il primo elemento
      per gli elementi di posto pari e dividendo per gli elementi di posto dispari.
      Se trascendente è nullo, ritorna una lista vuota.
      Usa: PrimoSuccessivo(x)
      Parametri:
          ripetizioni, numero di iterazioni
          trascendente, numero irrazionale o trascendente da approssimare
      Valore restituito:
          una lista con il segno del numero da approssimare e coppie
          di numeri primi che moltiplicati danno l'approssimazione
    """

    if trascendente == 0:
      return []
    segno = 1
    # Se il numero da approssimare è negativo, conserva il
    # segno e approssima il corrispondente numero positivo.
    if trascendente < 0:
      trascendente = -trascendente
      segno = -1
    p2 = 1
    t = 1
    l = [segno] * 1
    for count4 in range(int(ripetizioni)):
      p1 = PrimoSuccessivo(p2)
      p2 = PrimoSuccessivo(p1)
      # Se l'approssimazione (t) è minore del numero da approssimare
      # la moltiplica per il rapporto p2/p1 (che è maggiore di uno),
      # altrimenti per il rapporto p1/p2 (che è minore di uno).
      if t < trascendente:
        l.append(p2)
        l.append(p1)
        t = t * (p2 / p1)
      elif t > trascendente:
        l.append(p1)
        l.append(p2)
        t = t * (p1 / p2)
    return l


def Taylor_atan(iterazioni, x):
  """
    Approssima l'arcotangente di x sviluppandola in serie di Taylor in un intorno di zero.
    Lo sviluppo è del tipo x - x^3/3 + x^5/5 - ... Iterazioni è il numero di iterazioni (il grado del polinomio).
    Se iterazioni è minore di uno restituisce infinito.
      Usa: ...
      Parametri:
          iterazioni, il numero di passi
          x, il punto in cui calcolare la funzione mediante l'approssimazione
      Valore restituito
          f, l'approssimazione di atan(x)
  """

  # Controlla che il iterazioni sia maggiore o uguale
  # a uno, in caso contrario restituisce infinito
  if iterazioni < 1:
    return float('inf')
  iterazioni = round(iterazioni)
  f = x
  # Aggiunge allo sviluppo i termini del tipo -x^3/3 ... per gli indici 3, 7, 11, ...
  for i in (3 <= float(iterazioni)) and upRange(3, float(iterazioni), 4) or downRange(3, float(iterazioni), 4):
    f = f - x ** i / i
  # Aggiunge allo sviluppo i termini del tipo +x^5/5 ... per gli indici 5, 9, 13, ...
  for j in (5 <= float(iterazioni)) and upRange(5, float(iterazioni), 4) or downRange(5, float(iterazioni), 4):
    f = f + x ** j / j
  return f


def Taylor_PiAtan1(iterazioni):
  """
    Approssima pi greco come 4 * atan(1). Se iterazioni è minore di uno restituisce infinito.
    Usa: Taylor_atan
    Parametri:
      iterazioni, il numero di iterazioni
    Valore restituito:
      un numero decimale che approssima pi greco
  """

  x = 4 * Taylor_atan(iterazioni, 1)
  return x


def Taylor_PiAtan1_2_1_3(iterazioni):
  """
    Approssima pi greco come 4 * (atan(1/2)+atan(1/3)).
    Se iterazioni è minore di uno restituisce infinito.
    Usa: Taylor_atan
    Parametri:
        iterazioni, il numero di iterazioni
    Valore restituito:
        un numero decimale che approssima pi greco
  """

  p1 = Taylor_atan(iterazioni, 1 / 2)
  p2 = Taylor_atan(iterazioni, 1 / 3)
  return 4 * (p1 + p2)


def Taylor_e(iterazioni):
  """
    Approssima il numero di Eulero e utilizzando lo sviluppo in serie di Taylor della funzione e^x
    calcolata per x=1. Iterazioni deve essere un numero intero maggiore di zero, altrimenti la funzione
    restituisce infinito.
    Usa: Fattoriale
    Parametri:
      iterazioni: il numero di termini della serie usata per approssimare
    Valore restituito:
      un'approssimazione di e se iterazioni è un numero intero e positivo, infinito in caso contrario.
  """

  # Controlla che il iterazioni sia maggiore o uguale
  # a uno, in caso contrario restituisce infinito
  if iterazioni < 1:
    return float('inf')
  iterazioni = round(iterazioni)
  p = 1
  for l in (1 <= float(iterazioni)) and upRange(1, float(iterazioni), 1) or downRange(1, float(iterazioni), 1):
    p = p + 1 / Fattoriale(l)
  return p


def ld(x):
  """
    Calcola la derivata logaritmica di un numero intero.
    La derivata logaritmica è uguale alla somma
    dei reciproci dei fattori primi del numero.
      Usa: ScomporreInFattori
      Parametri:
        x, numero intero
      Valore restituito:
        la derivata logaritmica di x
  """

  # Se il numero specificato (x) è 0 oppure uno la derivata
  # logaritmica è nulla, per cui restituisce immediatamente 0.
  if x == 0 or x == 1:
    return 0
  # Se il numero specificato (x) è negativo lo trasforma in
  # positivo moltiplicandolo per -1, sfrutta la proprietà
  # della derivata logaritmica per cui ld(n) = ld(-n).
  # In questo modo alla funzione ScomporreInFattori è
  # passato un valore positivo (diverso da zero e da
  # uno) in modo che possa funzionare correttamente.
  if x < 0:
    x = -1 * x
  # Scompone in fattori il numero specificato chiamando
  # la funzione ScomporreInFattori. La variabile ld
  # contiene ora la lista dei fattori primi di x.
  lderivative = ScomporreInFattori(x)
  somma = 0
  # Calcola la somma dei reciproci dei fattori primi
  # di x (e quindi la sua derivata logaritmica).
  for p in lderivative:
    somma = 1 / p + somma
  return somma


def DAIntero(x):
  """
    Calcola la derivata aritmetica di un numero intero. Usa la funzione ld(x) (derivata logaritmica
    di un intero). La derivata aritmetica è uguale al numero specificato per la derivata logaritmica
    dello stesso numero.
    Usa: ...
    Parametri:
      x un numero intero
    Valore restituito:
      la derivata aritmetica di x
  """

  # ld è la funzione che calcola la derivata logaritmica del numero dato.
  somma = ld(x)
  return x * somma


def DARazionale(numeratore, denominatore):
  """
    Calcola la derivata aritmetica del numero razionale specificato considerando tutti i fattori del numeratore
    con esponente +1 e tutti i fattori del denominatore con esponente -1.
    Usa: ld (derivata logaritmica)
    Parametri:
        numeratore, numeratore del numero da derivare
        denominatore, denominatore del numero da derivare
    Valore restituito:
        derivata aritmetica del numero razionale specificato
  """

  # Calcola la derivata logaritmica del numeratore.
  a = ld(numeratore)
  # Calcola la derivata logaritmica del denominatore.
  b = ld(denominatore)
  return (numeratore / denominatore) * (a - b)


def DAReale(x, iterazioni):
  """
    Calcola la derivata aritmetica del numero reale specificato: il numero è sviluppato con il metodo del minimo
    scostamento in un prodotto di primi con esponente +1 o -1, la derivata è calcolata sommando i reciproci
    dei fattori del prodotto moltiplicati per il corrispondente esponente.
    Se x è intero è restituita la derivata di x come intero.
    Usa: MinimoScostamentoLista
    Parametri:
      x, numeratore reale da derivare
      iterazioni, numero dei passi dello sviluppo con il minimo scostamento
    Valore restituito:
      approssimazione della derivata aritmetica del numero reale specificato
  """

  if x % 1 == 0:
    return DAIntero(x)
  segno = 1
  # Memorizza il segno del numero da derivare. La derivata
  # del valore assoluto del numero da derivare sarà
  # moltiplicata per il segno prima di restituire il risultato.
  if x < 0:
    segno = -1
  # Sviluppa il valore assoluto numero da derivare
  # usando il metodo del minimo scostamento
  l = MinimoScostamentoLista(iterazioni, math.fabs(x))
  somma = l.pop(0)
  somma = 0
  while not not len(l):
    somma = somma + 1 / l.pop(0)
    somma = somma - 1 / l.pop(0)
  return segno * (x * somma)


